# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging.config
import os
import sys

import yaml
from dotenv import load_dotenv

import ondewo.logging.__init__ as file_anchor

load_dotenv()
MODULE_NAME = os.getenv("MODULE_NAME", "")


def import_config():  # type: ignore
    """
    Imports the config from the yaml file. The yaml file is taken relative to this file, so nothing about the python path is assumed.

    :param:
    :return:    logging config as a dictionary
    """
    if os.path.exists("./logging.yaml"):
        config_path = "./logging.yaml"
    else:
        parent = os.path.abspath(os.path.dirname(file_anchor.__file__))
        config_path = "{}/config/logging.yaml".format(parent)

    with open(config_path) as fd:
        conf = yaml.safe_load(fd)

    return conf  # type: ignore


def set_module_name(module_name, conf):  # type: ignore
    """
    Sets the module name as given in the env on deployment.

    :param module_name:         the name of the module the logger is running in
    :param conf:                the config of the logger
    :return:                    the config with module name
    """
    conf["logging"]["formatters"]["fluent_console"]["format"][
        "module_name"
    ] = module_name
    conf["logging"]["formatters"]["fluent_debug"]["format"]["module_name"] = module_name

    return conf


def initiate_loggers(conf):  # type: ignore
    """
    Initiate the loggers with the config and return them. Will complain if the module name is not set.

    :param conf:                the config of the logger
    :return:                    the loggers
    """
    logging.config.dictConfig(conf["logging"])

    if not MODULE_NAME:
        logging.warning(
            "NO MODULE NAME WAS GIVEN please give a module name for the fluent logs"
        )

    logger_root = logging.getLogger("root")
    logger_console = logging.getLogger("console")
    logger_debug = logging.getLogger("debug")

    logger = logger_root

    return logger, logger_root, logger_debug, logger_console


def check_python_version():  # type: ignore
    """
    Checks the python version. The python3 logger cant be imported to python 2.

    :param:
    :return:
    """
    if sys.version_info[0] == 3:
        logger_console.error(
            "Looks like you imported the Python2 logger in a Python3 project. Did you mean to do that?"
        )


conf = import_config()
conf = set_module_name(MODULE_NAME, conf)
logger, logger_root, logger_debug, logger_console = initiate_loggers(conf)
check_python_version()



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import time
import traceback
import uuid
from contextlib import ContextDecorator
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Optional, Union

from ondewo.logging.constants import CONTEXT, EXCEPTION, FINISH, START
from ondewo.logging.logger import logger_console


@dataclass
class Timer(ContextDecorator):
    """Time your code using a class, context manager, or decorator"""

    name: Optional[str] = None
    message: str = FINISH
    logger: Optional[
        Callable[[Union[str, Dict[str, Any]]], None]
    ] = logger_console.warning
    _start_time: Optional[float] = field(default=None, init=False, repr=False)
    log_arguments: bool = True
    suppress_exceptions: bool = False
    recursive: bool = False
    recurse_depth: int = 0
    argument_max_length: int = 10000

    def __post_init__(self) -> None:
        """Initialization: add timer to dicCallable[[Union[str, Dict[str, Any]]], None]t of timers"""
        if not self.name:
            self.name = str(uuid.uuid4())

    def __call__(self, func: Callable) -> Callable:  # type: ignore
        """
        Decorator which adds a logs timing information for the decorated function.

        :param func:    the function to be decorated
        :return:        the decorator
        """

        @functools.wraps(func)
        def wrapper_timing(*args, **kwargs) -> Any:  # type: ignore
            self.start(func)

            try:
                value: Any = func(*args, **kwargs)
            except Exception as exc:
                trace = traceback.format_exc()
                log_exception(type(exc), exc.args[0], trace, func.__name__, self.logger)  # type: ignore
                if not self.suppress_exceptions:
                    self.stop(func.__name__)
                    raise
                value = "An exception occurred!"

            if self.log_arguments:
                log_args_kwargs_results(
                    func, value, self.argument_max_length, self.logger, *args, **kwargs
                )

            self.stop(func.__name__)
            return value

        return wrapper_timing

    def start(self, func: Optional[Callable] = None) -> None:
        """Start a new timer"""
        if func:
            self.logger({"message": START.format(func.__name__)})  # type: ignore
        if self._start_time is not None:
            if self.recursive:
                self.recurse_depth += 1
                self.logger(f"Recursing, depth = {self.recurse_depth}")  # type: ignore
                return
        else:
            self._start_time = time.perf_counter()

    def stop(self, func_name: Optional[str] = None) -> float:
        """Stop the timer, and report the elapsed time"""
        if self.recurse_depth:
            self.recurse_depth -= 1
            if self.recursive:
                return 0.0

        # Calculate elapsed time
        assert self._start_time
        elapsed_time = time.perf_counter() - self._start_time

        # Report elapsed time
        if self.logger:
            self.report(elapsed_time, func_name)

        return elapsed_time

    def report(self, elapsed_time: float, func_name: Optional[str]) -> None:
        name = func_name if func_name else CONTEXT

        try:
            message = self.message.format(elapsed_time, name)
        except IndexError:
            pass

        log: Dict[str, Any] = {
            "message": message,
            "duration": elapsed_time,
            "tags": ["timing"],
        }
        self.logger(log)  # type: ignore

    def __enter__(self) -> "Timer":
        """Start a new timer as a context manager"""
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: str, traceback_obj: Any) -> bool:
        """Stop the context manager timer"""
        self.stop()
        if exc_type:
            log_exception(
                exc_type, exc_val, traceback.format_exc(), CONTEXT, self.logger  # type: ignore
            )
        return self.suppress_exceptions


timing = Timer()


def exception_silencing(func: Callable) -> Callable:
    """
    Quietly kills exceptions, logging the minimum.

    :param func:    the function to be decorated
    :return:        the decorator

    """

    @functools.wraps(func)
    def wrapper_error_silent(*args, **kwargs) -> Any:  # type: ignore
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            log_exception(type(exc), next(iter(exc.args), None), None, func.__name__)
            log_args_kwargs_results(func, None, -1, None, *args, **kwargs)
        return None

    return wrapper_error_silent


def exception_handling(func: Callable) -> Callable:
    """
    Hands exceptions on to the logger. This is especially useful in threaded applications where exceptions are often swallowed by the threading demons.

    :param func:    the function to be decorated
    :return:        the decorator

    """

    @functools.wraps(func)
    def wrapper_error(*args, **kwargs) -> Any:  # type: ignore
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            trace = traceback.format_exc()
            log_exception(type(exc), next(iter(exc.args), None), trace, func.__name__)
            log_args_kwargs_results(func, None, -1, None, *args, **kwargs)
        return None

    return wrapper_error


def log_arguments(func: Callable) -> Callable:
    """
    decorate a function to log its arguments and result

    :param func:    the function to be decorated
    :return:        the decorator

    """

    @functools.wraps(func)
    def wrapper_args(*args, **kwargs) -> Any:  # type: ignore
        result = func(*args, **kwargs)
        log_args_kwargs_results(func, result, -1, None, *args, **kwargs)
        return result

    return wrapper_args


def log_exception(
    exc_type: Any,
    exc_val: Optional[str],
    traceback_str: Optional[str],
    function_name: str,
    logger: Callable[[Union[str, Dict[str, Any]]], None] = logger_console.error,
) -> None:
    """
    Formats and logs an exception.
    """
    message = EXCEPTION.format(exc_type, exc_val, function_name)
    log: Dict[str, Any] = {
        "message": message,
        "exception type": exc_type,
        "exception value": exc_val,
        "traceback": traceback_str,
        "tags": ["timing", "exception"],
    }
    logger(log)


def log_args_kwargs_results(  # type: ignore
    func: Callable,
    result: Any,
    argument_max_length: int = -1,
    logger: Optional[
        Callable[[Union[str, Dict[str, Any]]], None]
    ] = logger_console.warning,
    *args,
    **kwargs,
) -> None:
    def truncate_arg(arg_to_truncate: Any) -> str:
        arg_str: str = str(arg_to_truncate)
        if argument_max_length == -1:
            return arg_str
        if len(arg_str) > argument_max_length:
            return arg_str[:argument_max_length] + "<TRUNCATED!>"
        return arg_str

    """
    Format and log all the inputs and outputs of a function
    """
    args_to_log = {truncate_arg(arg) for arg in args}
    kwargs_to_log = {i: truncate_arg(kwargs[i]) for i in kwargs}
    formatted_results: Dict[str, Any] = {
        "function": f"{func.__name__}",
        "args": f"{args_to_log}",
        "kwargs": f"{kwargs_to_log}",
        **kwargs_to_log,
        "result": f"{truncate_arg(result)}",
    }
    if logger:
        logger(
            {
                "message": f"Function arguments log: {formatted_results}",
                **formatted_results,
            }
        )



START: str = "Starting {!r}!"
FINISH: str = "Elapsed time: {:0.4f} seconds. Finished {!r}"
CONTEXT: str = "ContextManager"

EXCEPTION: str = "An exception '{}' occurred, with message '{}'. Traceback is in debug log. Finished {!r}."


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requires = f.read().splitlines()

setuptools.setup(
    name="ondewo-logging",
    version="2.0.0",
    author="Ondewo GbmH",
    author_email="info@ondewo.com",
    description="This library provides custom logging for python including error handling and timing.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/ondewo/ondewo-logging",
    packages=[
        np
        for np in filter(
            lambda n: n.startswith('ondewo.') or n == 'ondewo',
            setuptools.find_packages()
        )
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires=">2.6, !=3.0.*, !=3.1.*",
    include_package_data=True,
    install_requires=requires,
)



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging.config
import os
import re
import sys
from typing import Any, Dict, Optional, Tuple

import yaml
from dotenv import load_dotenv

import ondewo.logging.__init__ as file_anchor

load_dotenv()
MODULE_NAME: str = os.getenv("MODULE_NAME", "")
GIT_REPO_NAME: str = os.getenv("GIT_REPO_NAME", "")
DOCKER_IMAGE_NAME: str = os.getenv("DOCKER_IMAGE_NAME", "")


def flatten_json(y: Any, max_level: int = 3) -> Dict:
    output = {}

    def flatten_recursively(node: Any, name: str = "", level: int = 0) -> None:
        assert max_level > 0
        if level < max_level:
            if type(node) is dict:
                for sub_node in node:
                    flatten_recursively(
                        node[sub_node], name + sub_node + "|", level + 1
                    )
            else:
                output[name[:-1]] = node
        else:
            if type(node) is dict:
                output[name[:-1]] = "<TRUNCATED!>"
            else:
                output[name[:-1]] = node

    flatten_recursively(y)
    return output


class CustomLogger(logging.Logger):
    GRPC_LEVEL_NUM = 25
    logging.addLevelName(GRPC_LEVEL_NUM, "GRPC")

    @staticmethod
    def extract_grpc_request_class(msg: str) -> str:
        pattern_request_type = re.compile("\\(type <class '(.*)'>")
        request_type_search_result = pattern_request_type.search(msg)
        assert request_type_search_result
        request_type = request_type_search_result.group(1)
        return request_type

    @staticmethod
    def extract_grpc_message(msg: str, kwargs) -> Dict:  # type: ignore
        extracted: Dict = {}
        grpc_matches = re.findall("{.*}", msg)
        if grpc_matches:
            longest_match = max(grpc_matches)
            grpc_json = json.loads(longest_match)
            if "max_level" in kwargs:
                max_level = kwargs["max_level"]
                extracted = flatten_json(grpc_json, max_level)
            else:
                extracted = flatten_json(grpc_json)
        return extracted

    def grpc(self, message_dict: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore
        if logging.getLogger().isEnabledFor(self.GRPC_LEVEL_NUM):
            message: str = message_dict["message"]
            if "Got request (type <class" in message:
                request_type = CustomLogger.extract_grpc_request_class(message)
                extracted = CustomLogger.extract_grpc_message(message, kwargs)
                to_log = {"original": message, **extracted}
                # Add request type
                if request_type:
                    to_log["request_type"] = request_type
                    # Update tags with 'grpc'
                    to_log["tags"] = (
                        message_dict["tags"] + ["grpc"]
                        if "tags" in message_dict
                        else ["grpc"]
                    )
                if "max_level" in kwargs:
                    kwargs.pop("max_level")
                return super(CustomLogger, self).log(
                    self.GRPC_LEVEL_NUM, to_log, *args, **kwargs
                )
            else:
                if "max_level" in kwargs:
                    kwargs.pop("max_level")
                return super(CustomLogger, self).log(
                    self.GRPC_LEVEL_NUM, message_dict, *args, **kwargs
                )


def import_config() -> Dict[str, Any]:
    """
    Imports the config from the yaml file. The yaml file is taken relative to this file, so nothing about the python path is assumed.

    :param:
    :return:    logging config as a dictionary
    """
    if os.path.exists("./logging.yaml"):
        config_path: str = "./logging.yaml"
    else:
        parent: str = os.path.abspath(os.path.dirname(file_anchor.__file__))
        config_path = f"{parent}/config/logging.yaml"

    with open(config_path) as fd:
        conf = yaml.safe_load(fd)

    return conf  # type: ignore


def set_module_name(
    module_name: str, git_repo_name: str, docker_image_name: str, conf: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Sets the module name as given in the env on deployment.

    :param module_name:         the name of the module the logger is running in
    :param conf:                the config of the logger
    :return:                    the config with module name
    """
    conf["logging"]["formatters"]["fluent_console"]["format"][
        "module_name"
    ] = module_name
    conf["logging"]["formatters"]["fluent_debug"]["format"]["module_name"] = module_name
    conf["logging"]["formatters"]["fluent_console"]["format"][
        "git_repo_name"
    ] = git_repo_name
    conf["logging"]["formatters"]["fluent_debug"]["format"][
        "git_repo_name"
    ] = git_repo_name
    conf["logging"]["formatters"]["fluent_console"]["format"][
        "docker_image_name"
    ] = docker_image_name
    conf["logging"]["formatters"]["fluent_debug"]["format"][
        "docker_image_name"
    ] = docker_image_name

    return conf


def initiate_loggers(conf: Dict[str, Any]) -> Tuple[logging.Logger, ...]:
    """
    Initiate the loggers with the config and return them. Will complain if the module name is not set.

    :param conf:                the config of the logger
    :return:                    the loggers
    """
    logging.setLoggerClass(CustomLogger)
    logging.config.dictConfig(conf["logging"])

    if not GIT_REPO_NAME:
        logging.warning(
            "NO GIT REPO NAME WAS GIVEN please set the GIT_REPO_NAME env var\n"
            "   the GIT REPO NAME provides information to the logs\n"
            "   that may be unavailable in the image\n"
            "   it can (and should) be set at the start of the Dockerfile\n"
        )
    if not DOCKER_IMAGE_NAME:
        logging.warning(
            "NO DOCKER IMAGE NAME WAS GIVEN please state the docker image for the fluent logs\n"
        )
    if not MODULE_NAME:
        logging.warning(
            "NO MODULE NAME WAS GIVEN please give a module name for the fluent logs\n"
            "   the MODULE NAME providers information about the environment\n"
            "   such as which deployment the logs are coming from\n"
        )
    if not os.path.exists("./logging.yaml"):
        logging.info(
            "No logging.yaml in the root of the project, using the default config.\n"
        )

    logger_root: logging.Logger = logging.getLogger("root")
    logger_console: logging.Logger = logging.getLogger("console")
    logger_debug: logging.Logger = logging.getLogger("debug")

    logger: logging.Logger = logger_root

    return logger, logger_root, logger_debug, logger_console


def check_python_version(logger: logging.Logger) -> None:
    """
    Checks the python version. The python3 logger cant be imported to python 2.

    :param:
    :return:
    """
    if sys.version_info[0] == 2:
        # this wont actually run because of syntax errors, but functions as a sort of documentation
        logger.error(
            "Looks like you imported the Python3 logger in a Python2 project. Did you mean to do that?"
        )


def create_logs(conf: Optional[Dict[str, Any]] = None) -> Tuple[logging.Logger, ...]:
    """
    Loads, configures and returns logs.

    :param conf:    configuration dictionary, so the loggers can be configured in individual submodules.
    :return:        all the loggers
    """
    conf = conf if conf else import_config()
    conf = set_module_name(MODULE_NAME, GIT_REPO_NAME, DOCKER_IMAGE_NAME, conf)
    logger, logger_root, logger_debug, logger_console = initiate_loggers(conf)
    check_python_version(logger_console)
    return logger, logger_root, logger_debug, logger_console


logger, logger_root, logger_debug, logger_console = create_logs()
